@@ -21,8 +21,8 @@ classes, like handlers for a kernel module, a driver package, a handler group,
 etc.
 
 Custom concrete handlers need to fulfill the following requirements:
- - __init__(self, ui) must take exactly one argument (a reference to an
-   AbstractUI instance). All othe properties must be detected by the
+ - __init__(self, backend) must take exactly one argument (a reference to a
+   Backend instance). All othe properties must be detected by the
    constructor or changed with methods. These classes are instantiated
    automatically, which is not possible with constructors which need more
    arguments.
@@ -42,28 +42,63 @@ from oslib import OSLib
 class Handler:
     '''Abstract basic handler.'''
 
-    def __init__(self, ui, name, description=None, rationale=None):
+    def __init__(self, backend, name, description=None, rationale=None):
         '''Create a handler with given (human readable) name.
         
         Every handler should have a human readable name. A custom rationale and
         a multi-line description can be given, too. Every handler gets a
-        reference to the currently used user interface (subclass of AbstractUI)
-        so that it can make callbacks and query the user.
+        reference to the currently used Backend so that it can request
+        installation of packages and other system changes.
         '''
+        self._hwids = [] # covered HardwareIDs
         self._changed = False
+        self.backend = backend
+
         self._name = name
         self._description = description
-        self.ui = ui
         self._rationale = rationale
+        self.license = None
+
+        # the following properties are not specified in the ctor, since they
+        # might be changed after instantiation;
+        # subclass ctors might set that before calling us
+        if not hasattr(self, '_free'):
+            self._free = None
+        if hasattr(self, 'package') and self.package:
+            self._package_defaults()
+        else:
+            self.package = None
+        self.driver_vendor = None
+        self.version = None
+        self.repository = None
+        self._recommended = False
+
+    def _package_defaults(self):
+        '''Set fallback name/description/freeness from package.'''
+
+        if self.package and (not self._name or self._description is None):
+            (distro_name, distro_desc) = OSLib.inst.package_description(self.package)
+            if not self._name:
+                self._name = distro_name
+            if self._description is None:
+                self._description = distro_desc
+        if self.package and self._free is None:
+            try:
+                self._free = OSLib.inst.is_package_free(self.package)
+            except (KeyError, ValueError):
+                # we cannot determine it right now
+                pass
 
     def name(self):
         '''Return one-line name of the handler (for human consumption).'''
 
+        self._package_defaults()
         return self._name
 
     def description(self):
         '''Return multi-line description of the handler.'''
 
+        self._package_defaults()
         return self._description
 
     def id(self):
@@ -72,24 +107,20 @@ class Handler:
         This is used for specifying a handler with --enable/--disable on the
         command line, and is mentioned in the --list output.
         '''
-        return '%s:%s' % (str(self.__class__).split('.')[-1], self.name())
+        if self.package:
+            i = 'pkg:' + self.package
+        else:
+            i = '%s:%s' % (str(self.__class__).split('.')[-1], self.name())
+        if self.driver_vendor:
+            i += ':' + self.driver_vendor.replace(' ', '_')
+        return i
 
     def rationale(self):
         '''Return rationale as to why this driver might be enabled.
         
         Might return None if no rationale is available.
         '''
-        if self._rationale:
-            return self._rationale
-
-        # default rationale
-        if self.free():
-            return None
-        else:
-            return _('This driver is necessary to support the '
-                'hardware, there is no free/open alternative.\n\n'
-                'If this driver is not enabled, the hardware will not '
-                'function.')
+        return self._rationale
 
     def changed(self):
         '''Return if the module has been enabled/disabled at least once.'''
@@ -124,7 +155,10 @@ class Handler:
     def free(self):
         '''Return if the handler represents a free software driver.'''
 
-        raise NotImplementedError, 'subclasses need to implement this'
+        if self._free is not None:
+            return self._free
+        else:
+            raise NotImplementedError, 'subclasses need to implement this'
 
     def enabled(self):
         '''Return if the handler is enabled.
@@ -132,13 +166,26 @@ class Handler:
         'Enabled' means that the user agreed to use this driver if it is
         applicable.
         '''
-        raise NotImplementedError, 'subclasses need to implement this'
+        if self.package:
+            return OSLib.inst.package_installed(self.package)
+        else:
+            return True
 
     def used(self):
         '''Return if the handler is currently in use.'''
 
         raise NotImplementedError, 'subclasses need to implement this'
 
+    def recommended(self):
+        '''Return if the version of a certain driver is recommended over others
+        when more than one driver flavour supports the same device.
+        
+        This method should return True only for the recommended version while
+        it will return False for any other compatible version. If only one
+        version of a driver is provided, then it should return False.
+        '''
+        return self._recommended
+    
     def available(self):
         '''Return if the conditions to use this handler on the system are met.
 
@@ -151,27 +198,56 @@ class Handler:
         merely available in the handler pool, and an external driver database
         (detection.DriverDB) is queried.
         '''
-        raise NotImplementedError, 'subclasses need to implement this'
-
-    def ui_category(self):
-        '''Return handler category (translated, human readable, short).'''
-
-        raise NotImplementedError, 'subclasses need to implement this'
+        if self.package:
+            if not self.repository or OSLib.inst.repository_enabled(self.repository):
+                try:
+                    OSLib.inst.package_description(self.package)
+                    return None
+                except ValueError:
+                    return False
+            else:
+                return None # undecidable until the repo is added
+        else:
+            raise NotImplementedError, 'subclasses need to implement this'
 
     def enable(self):
         '''Allow the OS to use it if the hardware is available.
         
-        If possible, the handler should be loaded, too.
+        If possible, the handler should be loaded, too. Return True if
+        immediately successful, or False if the system needs to be rebooted for
+        the changes to become effective.
         '''
+        if self.repository:
+            OSLib.inst.add_repository(self.repository)
+
+        if self.package:
+            self.backend.install_package(self.package)
+            if not OSLib.inst.package_installed(self.package):
+                # do not touch _changed if package failed to install
+                return True
         self._changed = True
+        return True
 
     def disable(self):
         '''Prevent the OS from using it even if the hardware is available.
 
-        If possible, the handler should be unloaded, too.
+        If possible, the handler should be unloaded, too. Return True if
+        immediately successful, or False if the system needs to be rebooted for
+        the changes to become effective.
+
         '''
+        if self.repository:
+            OSLib.inst.remove_repository(self.repository)
+
+        if self.package:
+            self.backend.remove_package(self.package)
+            if OSLib.inst.package_installed(self.package):
+                # do not touch _changed if package failed to remove
+                return
         self._changed = True
 
+        return True
+
 #--------------------------------------------------------------------#
 
 class HandlerGroup(Handler):
@@ -181,8 +257,8 @@ class HandlerGroup(Handler):
     similar handlers in the UI. For example, the three VMWare or the dozens of
     commercial OSS drivers should be grouped.
     '''
-    def __init__(self, ui, name, id, description=None, rationale=None):
-        Handler.__init__(self, ui, name, description, rationale)
+    def __init__(self, backend, name, id, description=None, rationale=None):
+        Handler.__init__(self, backend, name, description, rationale)
         self._id = id
         self.subhandlers = []
 
@@ -229,7 +305,7 @@ class HandlerGroup(Handler):
         If all subhandlers return False, this returns False. If any subhandler
         returns True, this returns True. Otherwise this returns None.
         '''
-        all_False = True
+        all_false = True
 
         for h in self.subhandlers:
             a = h.available()
@@ -245,22 +321,21 @@ class HandlerGroup(Handler):
         else:
             return None
 
-    def ui_category(self):
-        '''Return handler category (translated, human readable, short).'''
-
-        return self.subhandlers[0].ui_category()
-
     def enable(self):
         '''Enable all subhandlers.'''
 
+        result = True
         for h in self.subhandlers:
-            h.enable()
+            result = h.enable() and result
+        return result
 
     def disable(self):
         '''Disable all subhandlers.'''
 
+        result = True
         for h in self.subhandlers:
-            h.disable()
+            result = h.disable() and result
+        return result
 
     def changed(self):
         '''Return if at least one subhandler has been enabled/disabled at
@@ -295,23 +370,26 @@ class KernelModuleHandler(Handler):
     '''
     _loaded_modules = None
     
-    def __init__(self, ui, module, name=None, description=None, rationale=None):
+    def __init__(self, backend, kernel_module, name=None, description=None, rationale=None):
         '''Create handler for a kernel module.
         
         If not given explicitly, the name is read from modinfo's 'description'
         field.
         '''
-        self.module = module
+        self.module = kernel_module
         self._modinfo = detection.get_modinfo(self.module)
-        assert self._modinfo, 'kernel module %s exists' % self.module
         if not name:
+            assert self._modinfo, 'kernel module %s exists' % self.module
             name = '\n'.join(self._modinfo.get('description', [self.module]))
-        Handler.__init__(self, ui, name, description, rationale)
+        Handler.__init__(self, backend, name, description, rationale)
 
     def id(self):
         '''Return an unique identifier of the handler.'''
 
-        return 'kmod:' + self.module
+        i = 'kmod:' + self.module
+        if self.driver_vendor:
+            i += ':' + self.driver_vendor.replace(' ', '_')
+        return i
 
     def free(self):
         '''Return if the handler represents a free software driver.'''
@@ -319,11 +397,14 @@ class KernelModuleHandler(Handler):
         # this function needs to be kept in sync with the kernel function
         # is_license_gpl_compatible()
 
+        if self._free is not None:
+            return self._free
+
+        assert self._modinfo, 'kernel module %s exists' % self.module
         for l in self._modinfo.get('license', ['unknown']):
             if l in ('GPL', 'GPL v2', 'GPL and additional rights', 
-                'Dual BSD/GPL', 'Dual MIT/GPL', 'Dual MPL/GPL'):
+                'Dual BSD/GPL', 'Dual MIT/GPL', 'Dual MPL/GPL', 'BSD'):
                 return True
-
         return False
 
     def enabled(self):
@@ -332,12 +413,14 @@ class KernelModuleHandler(Handler):
         'Enabled' means that the user agreed to use this driver if it is
         applicable.
         '''
-        return not OSLib.inst.module_blacklisted(self.module)
+        return not OSLib.inst.module_blacklisted(self.module) and \
+            (self._modinfo is not None) and Handler.enabled(self)
 
     def used(self):
         '''Return if the handler is currently in use.'''
 
-        return self.module_loaded(self.module)
+        return self.module_loaded(self.module) and (self.package is None or
+            OSLib.inst.package_installed(self.package))
 
     def available(self):
         '''Return if the conditions to use this handler on the system are met
@@ -348,11 +431,6 @@ class KernelModuleHandler(Handler):
         '''
         return None
 
-    def ui_category(self):
-        '''Return handler category (translated, human readable, short).'''
-
-        return self.ui._('Device driver')
-
     def enable(self):
         '''Allow the OS to use it if the hardware is available.
         
@@ -361,8 +439,9 @@ class KernelModuleHandler(Handler):
         Handler.enable(self)
         OSLib.inst.blacklist_module(self.module, False)
         subprocess.call([OSLib.inst.modprobe_path, self.module])
+        self._modinfo = detection.get_modinfo(self.module)
         self.read_loaded_modules()
-        self.rebind(self.module) # TODO: is this always wanted?
+        return self.rebind(self.module) # TODO: is this always wanted?
 
     def disable(self):
         '''Prevent the OS from using it even if the hardware is available.
@@ -371,12 +450,15 @@ class KernelModuleHandler(Handler):
         '''
         Handler.disable(self)
         OSLib.inst.blacklist_module(self.module, True)
+        self._modinfo = detection.get_modinfo(self.module)
+        return False # TODO: can we make this automatic?
 
     @classmethod
     def rebind(klass, module):
         '''Re-bind all devices using the module.
         
-        This is necessary for example to reload firmware.
+        This is necessary for example to reload firmware. Return True on
+        success, or False if rebind failed for any device.
         '''
         drivers_dir = os.path.join(OSLib.inst.sys_dir, 'module', module, 'drivers')
         if not os.path.isdir(drivers_dir):
@@ -384,6 +466,8 @@ class KernelModuleHandler(Handler):
                 drivers_dir, module))
             return
 
+        succeeded = True
+
         for driver in os.listdir(drivers_dir):
             driver_path = os.path.join(drivers_dir, driver)
             for device in os.listdir(driver_path):
@@ -398,6 +482,9 @@ class KernelModuleHandler(Handler):
                 except IOError:
                     logging.warning('unbind/rebind for device %s on driver %s failed', 
                         device, driver_path, exc_info=True)
+                    succeeded = False
+
+        return succeeded
 
     @classmethod
     def read_loaded_modules(klass):
@@ -419,7 +506,7 @@ class KernelModuleHandler(Handler):
 
     @classmethod
     def module_loaded(klass, module):
-        '''Return if the handler is currently in use.'''
+        '''Return if a module is currently loaded.'''
 
         if klass._loaded_modules == None:
             klass.read_loaded_modules()
@@ -436,7 +523,7 @@ class FirmwareHandler(KernelModuleHandler):
     This class' enable() function will deal with downloading it and the UI
     progress reporting of the download.
     '''
-    def __init__(self, ui, module, testfile, name=None, description=None, 
+    def __init__(self, backend, kernel_module, testfile, name=None, description=None, 
             rationale=None, url=None, sha1sum=None, free=False):
         '''Create handler for a piece of firmware for a kernel module.
         
@@ -449,7 +536,8 @@ class FirmwareHandler(KernelModuleHandler):
 
         By default this handler assumes that the firmware is not free (since
         otherwise the distribution could ship it together with the driver). You
-        can set 'free' to True for free firmware.
+        can set 'free' to True for free firmware or to None to use the kernel
+        module's freeness.
     
         If not given explicitly, the name is read from modinfo's 'description'
         field.
@@ -458,24 +546,24 @@ class FirmwareHandler(KernelModuleHandler):
         self.sha1sum = sha1sum
         self._free = free
         self.testfile = testfile
-        if not rationale:
-            if self._free:
-                rationale = ui.string_free_aux_firmware
-            else:
-                rationale = ui.string_nonfree_aux_firmware
 
-        KernelModuleHandler.__init__(self, ui, module, name, description,
-            rationale)
+        KernelModuleHandler.__init__(self, backend, kernel_module, name,
+            description, rationale)
+
+    def id(self):
+        '''Return an unique identifier of the handler.'''
+
+        i = 'firmware:' + self.module
+        if self.driver_vendor:
+            i += ':' + self.driver_vendor.replace(' ', '_')
+        return i
 
     def free(self):
         '''Return if the handler represents a free software driver.'''
 
-        if self._free == False:
-            return False
-        elif self._free is None:
+        if self._free is None:
             return KernelModuleHandler.free(self)
-        else:
-            return True
+        return self._free
 
     def enabled(self):
         '''Return if the handler is enabled.
@@ -490,20 +578,16 @@ class FirmwareHandler(KernelModuleHandler):
 
         return self.enabled() and KernelModuleHandler.used(self)
 
-    def ui_category(self):
-        '''Return handler category (translated, human readable, short).'''
-
-        return self.ui._('Firmware')
-
     def enable(self):
         '''Allow the OS to use it if the hardware is available.
         
         This downloads the url and puts it into self.firmware_file. Subclasses
         need to provide an actual implementation what to do with the file.
         '''
-        self.firmware_file = self.ui.download_url(self.url)[0]
-        if not self.firmware_file:
-            return
+        raise NotImplementedError, 'FirmwareHandler is currently not implemented'
+        #self.firmware_file = self.ui.download_url(self.url)[0]
+        #if not self.firmware_file:
+        #    return
 
         # TODO: sha1sum check
 
@@ -519,163 +603,27 @@ class FirmwareHandler(KernelModuleHandler):
 
 #--------------------------------------------------------------------#
 
-class DriverPackageHandler(Handler):
-    '''Handler for a driver package, not tied to a particular kernel module.
-    
-    Subclasses need to implement __init__(), available() and used() at least.
-    '''
-    def __init__(self, ui, package, name=None, description=None, 
-            rationale=None):
-        '''Create handler for of driver package.
-        
-        If not given explicitly, the name and description are taken from the
-        distribution package.
-        '''
-        self.package = package
-        if not name or not description:
-            (distro_name, distro_desc) = OSLib.inst.package_description(package)
-            if not name:
-                name = distro_name
-            if not description:
-                description = distro_desc
-
-        Handler.__init__(self, ui, name, description, rationale)
+class PrinterDriverHandler(Handler):
+    '''Handler for a printer driver.'''
 
     def id(self):
-        '''Return an unique identifier of the handler.'''
+        '''Return an unique identifier of the handler.
 
-        return 'pkg:' + self.package
-
-    def free(self):
-        '''Return if the handler represents a free software driver.'''
-
-        return OSLib.inst.is_package_free(self.package)
-
-    def enabled(self):
-        '''Return if the handler is enabled.
-        
-        'Enabled' means that the user agreed to use this driver if it is
-        applicable.
+        This is used for specifying a handler with --enable/--disable on the
+        command line, and is mentioned in the --list output.
         '''
-        return OSLib.inst.package_installed(self.package)
+        if self.package:
+            i = 'printer:' + self.package
+        else:
+            i = 'printer:%s' % self.name()
+        if self.version:
+            i += ':' + str(self.version)
+        elif self.driver_vendor:
+            i += ':' + self.driver_vendor.replace(' ', '_')
+        return i
 
-    def available(self):
-        '''Return if the conditions to use this handler on the system are met.
+    def used(self):
+        '''Return if the handler is currently in use.'''
 
-        This usually means that the hardware for this driver is available, but
-        there might be hardware independent drivers, too.
-        
-        This returns False if the package is not available, otherwise None.
-        Subclasses are welcome to add a positive detection, but should return
-        False if this method returns False, unless they require a separate
-        package repository.
-        '''
-        try:
-            OSLib.inst.package_description(self.package)
-            return None
-        except ValueError:
-            return False
-
-    def ui_category(self):
-        '''Return handler category (translated, human readable, short).'''
-
-        return self.ui._('Device driver')
-
-    def enable(self):
-        '''Allow the OS to use it if the hardware is available.
-        
-        If possible, the handler should be loaded, too.
-        '''
-        self.ui.install_package(self.package)
-        if OSLib.inst.package_installed(self.package):
-            Handler.enable(self)
-
-    def disable(self):
-        '''Prevent the OS from using it even if the hardware is available.
-
-        If possible, the handler should be unloaded, too.
-        '''
-        self.ui.remove_package(self.package)
-        Handler.disable(self)
-
-#--------------------------------------------------------------------#
-
-class ModulePackageHandler(KernelModuleHandler, DriverPackageHandler):
-    '''Kernel module which needs an additional driver package.
-    
-    Subclasses need to implement __init__(), available() and used() at least.
-    '''
-    def __init__(self, ui, module, package, name=None, description=None, 
-            rationale=None):
-        '''Handler for a kernel module which needs an additional driver package.
-        
-        If not given explicitly, the name is read from modinfo's 'description'
-        field.
-        '''
-        DriverPackageHandler.__init__(self, ui, package, name, description,
-            rationale)
-        KernelModuleHandler.__init__(self, ui, module, name, description,
-            rationale)
-
-    def id(self):
-        '''Return an unique identifier of the handler.'''
-
-        return KernelModuleHandler.id(self)
-
-    def can_change(self):
-        '''Check whether we can actually modify settings of this handler.
-
-        This might not be the case if e. g. the user manually modified a
-        configuration file. Return an explanatory text if settings can not be
-        changed, or None if changing is ok.
-        '''
-        return DriverPackageHandler.can_change(self) or \
-            KernelModuleHandler.can_change(self)
-
-    def free(self):
-        '''Return if the handler represents a free software driver.'''
-
-        return DriverPackageHandler.free(self) and \
-            KernelModuleHandler.free(self)
-
-    def available(self):
-        '''Return if the conditions to use this handler on the system are met.
-
-        This usually means that the hardware for this driver is available, but
-        there might be hardware independent drivers, too.
-        
-        This returns False if the package is not available, otherwise None.
-        Subclasses are welcome to add a positive detection, but should return
-        False if this method returns False.
-        '''
-        # KernelModuleHandler does not have an interesting implementation, so
-        # only query the package one by default
-        return DriverPackageHandler.available(self)
-
-    def enabled(self):
-        '''Return if the handler is enabled.
-        
-        'Enabled' means that the user agreed to use this driver if it is
-        applicable.
-        '''
-        dr = DriverPackageHandler.enabled(self)
-        km = KernelModuleHandler.enabled(self)
-        logging.debug('ModulePackageHandler(%s, %s): module enabled: %s, package installed: %s' % (
-            self.module, self.package, km, dr))
-        return dr and km
-
-    def enable(self):
-        '''Allow the OS to use it if the hardware is available.
-        
-        If possible, the handler should be loaded, too.
-        '''
-        DriverPackageHandler.enable(self)
-        KernelModuleHandler.enable(self)
-
-    def disable(self):
-        '''Prevent the OS from using it even if the hardware is available.
-
-        If possible, the handler should be unloaded, too.
-        '''
-        DriverPackageHandler.disable(self)
-        KernelModuleHandler.disable(self)
+        # TODO: query cups for actually using the driver
+        return self.enabled()
