@@ -18,18 +18,17 @@
 # with this program; if not, write to the Free Software Foundation, Inc.,
 # 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
-import unittest, sys, os, shutil, signal, time
+import unittest, sys, os, signal, time, traceback, urllib, tempfile
 from cStringIO import StringIO
 import SimpleHTTPServer, BaseHTTPServer
 
-from jockey.ui import AbstractUI
+import jockey.ui, jockey.handlers
 from jockey.oslib import OSLib
-from jockey.handlers import KernelModuleHandler, Handler
 from jockey.xorg_driver import XorgDriverHandler
 
 import sandbox
 
-class UITest(unittest.TestCase):
+class UITest(sandbox.LogTestCase):
     '''User interface tests.
 
     These save and restore sys.argv, so the individual tests are free to modify
@@ -54,8 +53,6 @@ class UITest(unittest.TestCase):
         f.write(self.big_file_contents)
         f.close()
 
-        self.open_app_log = os.path.join(OSLib.inst.workdir, 'open_app.log')
-
     def tearDown(self):
         sys.argv = self.orig_argv
         self.stop_capture()
@@ -66,18 +63,15 @@ class UITest(unittest.TestCase):
 
         # undo the effect of enabling/disabling handlers
         OSLib.inst._make_proc_modules()
-        KernelModuleHandler.read_loaded_modules()
+        OSLib.inst.reset_packages()
+        OSLib.inst._make_xorg_conf()
+        jockey.handlers.KernelModuleHandler.read_loaded_modules()
         try:
             os.unlink(OSLib.inst.module_blacklist_file)
         except OSError:
             pass
         OSLib.inst._load_module_blacklist()
 
-        try:
-            os.unlink(self.open_app_log)
-        except OSError:
-            pass
-
         os.unlink(self.big_file)
 
     def stop_capture(self):
@@ -96,18 +90,21 @@ class UITest(unittest.TestCase):
         '''calling with --help'''
 
         # --help will exit, so we have to fork
+        self.stop_capture()
+        out = tempfile.TemporaryFile()
         if os.fork() == 0:
+            sys.stdout = out
             sys.argv = ['ui-test', '--help']
             try:
-                AbstractUI()
+                jockey.ui.AbstractUI()
             except SystemExit:
+                out.flush()
                 os._exit(0)
             os._exit(42) # we should not get here
 
         self.assertEqual(os.wait()[1], 0)
-
-        # TODO: check output (forked process does not land in self.stdout for
-        # some reason)
+        out.seek(0)
+        self.assert_('--update-db' in out.read())
 
     def test_list(self):
         '''calling with --list'''
@@ -115,7 +112,7 @@ class UITest(unittest.TestCase):
         open (os.path.join(OSLib.inst.handler_dir, 'kmod_nodetect.py'), 'w').write(
             sandbox.h_availability_py)
         sys.argv = ['ui-test', '--list']
-        self.assertEqual(AbstractUI().run(), 0)
+        self.assertEqual(sandbox.TestUI().run(), 0)
 
         self.stop_capture()
 
@@ -126,6 +123,34 @@ class UITest(unittest.TestCase):
         self.assert_('kmod:vanilla3d ' in self.stdout)
         self.assert_('kmod:firmwifi ' in self.stdout)
 
+    def test_hwids(self):
+        '''calling with --hardware-ids'''
+
+        sys.argv = ['ui-test', '--hardware-ids']
+        self.assertEqual(sandbox.TestUI().run(), 0)
+        self.stop_capture()
+
+        # filter out printers, since we cannot predict them
+        lines = [l for l in self.stdout.strip().splitlines() 
+            if not l.startswith('printer_deviceid:')]
+
+        self.assertEqual(len(lines), 7)
+        self.assert_('modalias:fire:1.2' in lines)
+
+    def test_update(self):
+        '''calling with --update'''
+
+        sys.argv = ['ui-test', '--update']
+        ui = sandbox.TestUI()
+        ui.backend().driver_dbs.append(sandbox.TestDriverDB())
+        self.assertEqual(ui.run(), 0)
+
+        self.stop_capture()
+
+        handlers = ui.backend().available()
+        self.assert_('kmod:vanilla3d' in handlers) # from LocalKMod
+        self.assert_('kmod:spam' in handlers) # from TestDriverDB
+
     def test_disable(self):
         '''calling with --disable'''
 
@@ -137,11 +162,9 @@ class UITest(unittest.TestCase):
         self.assertEqual(ui.run(), 0)
         self.stop_capture()
 
-        for h in ui.handlers:
-            if h.id() == 'kmod:mint':
-                self.failIf(h.enabled())
-            else:
-                self.assert_(h.enabled())
+        for h in ui.backend().available():
+            en = jockey.ui.bool(ui.backend().handler_info(h)['enabled'])
+            self.assertEqual(en, h != 'kmod:mint')
 
     def test_disable_nochange(self):
         '''calling with --disable on a non-changeable handler'''
@@ -154,8 +177,8 @@ class UITest(unittest.TestCase):
         self.assertNotEqual(ui.run(), 0)
         self.stop_capture()
 
-        for h in ui.handlers:
-            self.assert_(h.enabled())
+        for h in ui.backend().available():
+            self.assert_(jockey.ui.bool(ui.backend().handler_info(h)['enabled']))
 
         self.assert_('I must live' in self.stderr)
 
@@ -171,8 +194,8 @@ class UITest(unittest.TestCase):
         self.assertEqual(ui.run(), 1)
         self.stop_capture()
 
-        for h in ui.handlers:
-            self.assert_(h.enabled())
+        for h in ui.backend().available():
+            self.assert_(jockey.ui.bool(ui.backend().handler_info(h)['enabled']))
 
     def test_disable_confirm(self):
         '''calling with --disable --confirm OKing'''
@@ -186,11 +209,9 @@ class UITest(unittest.TestCase):
         self.assertEqual(ui.run(), 0)
         self.stop_capture()
 
-        for h in ui.handlers:
-            if h.id() == 'kmod:mint':
-                self.failIf(h.enabled())
-            else:
-                self.assert_(h.enabled())
+        for h in ui.backend().available():
+            en = jockey.ui.bool(ui.backend().handler_info(h)['enabled'])
+            self.assertEqual(en, h != 'kmod:mint')
 
     def test_enable(self):
         '''calling with --enable'''
@@ -201,10 +222,7 @@ class UITest(unittest.TestCase):
         # disable mint driver and verify that in --list
         sys.argv = ['ui-test', '--list']
         ui = sandbox.TestUI()
-        for h in ui.handlers:
-            if h.id() == 'kmod:mint':
-                h.disable()
-                break
+        ui.backend().set_enabled('kmod:mint', False)
         self.assertEqual(ui.run(), 0)
 
         # enable it using the CLI
@@ -222,8 +240,8 @@ class UITest(unittest.TestCase):
                 break
         self.assert_(mint_disabled)
 
-        for h in ui.handlers:
-            self.assert_(h.enabled())
+        for h in ui.backend().available():
+            self.assert_(jockey.ui.bool(ui.backend().handler_info(h)['enabled']))
 
     def test_enable_invalid(self):
         '''calling with --enable on a nonexisting handler'''
@@ -235,45 +253,8 @@ class UITest(unittest.TestCase):
         self.assert_(ui.string_unknown_driver in self.stderr)
         self.assert_('--list' in self.stderr)
 
-    def test_handler_dir(self):
-        '''calling with --handler-dir'''
-
-        mydir = os.path.join(OSLib.inst.workdir, 'customdir')
-        try:
-            os.mkdir(mydir)
-            open (os.path.join(mydir, 'kmod_nodetect.py'), 'w').write(
-                sandbox.h_availability_py)
-            sys.argv = ['ui-test', '--list', '-H', mydir]
-            self.assertEqual(AbstractUI().run(), 0)
-
-            self.stop_capture()
-
-            self.assertEqual(len(self.stdout.strip().splitlines()), 5)
-            self.assert_('free module with available hardware, graphics card' in self.stdout)
-            self.assert_('kmod:mint ' in self.stdout)
-            self.assert_('kmod:vanilla ' in self.stdout)
-            self.assert_('kmod:vanilla3d ' in self.stdout)
-            self.assert_('kmod:firmwifi ' in self.stdout)
-        finally:
-            shutil.rmtree(mydir)
-
-    def test_get_handler_tooltip(self):
-        '''get_handler_tooltip()'''
-
-        sys.argv = ['ui-test']
-        ui = AbstractUI()
-
-        self.assertEqual(ui.get_handler_tooltip(None), None)
-
-        h = KernelModuleHandler(ui, 'vanilla')
-        self.assertEqual(ui.get_handler_tooltip(h), None)
-
-        r = 'I am important'
-        h = Handler(ui, 'shiny driver', rationale=r)
-        self.assertEqual(ui.get_handler_tooltip(h), r)
-
-    def test_toggle_handler(self):
-        '''toggle_handler()'''
+    def test_set_handler_enable_error(self):
+        '''set_handler_enable() on errors'''
 
         open (os.path.join(OSLib.inst.handler_dir, 'h.py'), 'w').write(
             'import jockey.handlers' + sandbox.h_nochangemod)
@@ -282,48 +263,125 @@ class UITest(unittest.TestCase):
         ui = sandbox.TestUI()
         self.stop_capture()
 
-        for h in ui.handlers:
-            self.assert_(h.enabled())
-            if h.can_change():
-                no_change_h = h
+        for h_id in ui.backend().available():
+            self.assert_(jockey.ui.bool(ui.backend().handler_info(h_id)['enabled']))
+            if 'can_change' in ui.backend().handler_info(h_id):
+                no_change_h = h_id
             else:
-                change_h = h
-        assert no_change_h.can_change() != None
-        assert change_h.can_change() == None
+                change_h = h_id
+        assert no_change_h
+        assert change_h
 
         # change an unchangeable handler
         # note: don't set a confirmation response, it shouldn't ask here
-        self.assertEqual(ui.toggle_handler(no_change_h), False)
+        self.assertEqual(ui.set_handler_enable(no_change_h, 'enable', True), False)
+        self.assertEqual(ui.pop_error()[1], 'I must live')
+        self.assertRaises(IndexError, ui.pop_error)
+        self.assertEqual(ui.set_handler_enable(no_change_h, 'disable', True), False)
+        self.assertEqual(ui.pop_error()[1], 'I must live')
+        self.assertRaises(IndexError, ui.pop_error)
+        self.assertEqual(ui.set_handler_enable(no_change_h, 'toggle', True), False)
         self.assertEqual(ui.pop_error()[1], 'I must live')
         self.assertRaises(IndexError, ui.pop_error)
 
-        # do not confirm change of handler
-        ui.confirm_response = False
-        self.assertEqual(ui.toggle_handler(change_h), False)
-        self.assert_(change_h.enabled())
+        # invalid operation
+        self.assertRaises(ValueError, ui.set_handler_enable, no_change_h, 'foo', True)
+
+        # package install failure
+        ui.backend().handlers['kmod:mint'].package = 'pretzel'
+        self.assertEqual(jockey.ui.bool(ui.backend().handler_info('kmod:mint')['enabled']),
+            False)
+        OSLib.inst.pending_install_remove_exception = 'OMGbroken'
+        self.assertEqual(ui.set_handler_enable('kmod:mint', 'enable',
+            False), False)
+        self.assertEqual(jockey.ui.bool(ui.backend().handler_info('kmod:mint')['enabled']),
+            False)
+        # one error message about install failure
+        (etitle, etext) = ui.pop_error()
+        self.assert_('OMGbroken' in etext, '%s: %s' % (etitle, etext))
         self.assertRaises(IndexError, ui.pop_error)
 
-        # confirm change of handler
-        ui.confirm_response = True
-        self.assertEqual(ui.toggle_handler(change_h), True)
-        self.failIf(change_h.enabled())
+        # install it now
+        self.assertEqual(ui.set_handler_enable('kmod:mint', 'enable',
+            False), True)
+        self.assertEqual(jockey.ui.bool(ui.backend().handler_info('kmod:mint')['enabled']),
+            True)
+
+        # package removal failure
+        OSLib.inst.pending_install_remove_exception = 'OMGZombie'
+        self.assertEqual(ui.set_handler_enable('kmod:mint', 'disable',
+            False), False)
+        self.assertEqual(jockey.ui.bool(ui.backend().handler_info('kmod:mint')['enabled']),
+            True)
+        # one error message about removal failure
+        (etitle, etext) = ui.pop_error()
+        self.assert_('OMGZombie' in etext, '%s: %s' % (etitle, etext))
         self.assertRaises(IndexError, ui.pop_error)
 
-        # enable it again
-        ui.confirm_response = True
-        self.assertEqual(ui.toggle_handler(change_h), True)
-        self.assert_(change_h.enabled())
-        self.assertRaises(IndexError, ui.pop_error)
+    def test_set_handler_enable(self):
+        '''set_handler_enable()'''
+
+        sys.argv = ['ui-test']
+        ui = sandbox.TestUI()
+        self.stop_capture()
+
+        def case(confirm_response, action, confirm, expected_return, expected_enabled):
+            ui.confirm_response = confirm_response
+            self.assertEqual(ui.set_handler_enable('kmod:vanilla', action,
+                confirm), expected_return)
+            self.assertEqual(jockey.ui.bool(ui.backend().handler_info('kmod:vanilla')['enabled']),
+                expected_enabled)
+            self.assertRaises(IndexError, ui.pop_error)
+
+        # unconfirmed mode
+
+        # already enabled
+        case(None, 'enable', False, False, True)
+        # disabling changes something
+        case(None, 'disable', False, True, False)
+        # already disabled
+        case(None, 'disable', False, False, False)
+        # toggling always changes something
+        case(None, 'toggle', False, True, True)
+        case(None, 'toggle', False, True, False)
+
+        # confirmed mode
+
+        # already disabled, doesn't need confirm_respose
+        case(None, 'disable', True, False, False)
+        # enable, cancel confirmation
+        case(False, 'enable', True, False, False)
+        # enable, ack confirmation
+        case(True, 'enable', True, True, True)
+        # disable, cancel confirmation
+        case(False, 'disable', True, False, True)
+        # disable, ack confirmation
+        case(True, 'disable', True, True, False)
+        # toggle, cancel confirmation
+        case(False, 'toggle', True, False, False)
+        # toggle, ack confirmation
+        case(True, 'toggle', True, True, True)
+        case(True, 'toggle', True, True, False)
 
     def test_mode_any(self):
         '''calling with --mode any'''
 
-        sys.argv = ['ui-test', '--mode', 'any']
+        sys.argv = ['ui-test', '--list', '--mode', 'any']
         ui = sandbox.TestUI()
+        self.assertEqual(sandbox.TestUI().run(), 0)
+        displayed = set(ui.get_displayed_handlers())
+
         self.stop_capture()
 
-        self.assertEqual(set([h.module for h in ui.handlers]), 
-            set(['vanilla', 'vanilla3d', 'mint', 'firmwifi', 'foodmi']))
+        self.assertEqual(len(self.stdout.strip().splitlines()), 5)
+        self.assert_('free module with available hardware, graphics card' in self.stdout)
+        self.assert_('kmod:mint ' in self.stdout)
+        self.assert_('kmod:vanilla ' in self.stdout)
+        self.assert_('kmod:vanilla3d ' in self.stdout)
+        self.assert_('kmod:firmwifi ' in self.stdout)
+
+        self.assertEqual(displayed, set(['kmod:mint', 'kmod:vanilla',
+            'kmod:vanilla3d', 'kmod:firmwifi', 'kmod:foodmi']))
 
         # mint is not enabled by default
         self.failIf('Restricted' in ui.main_window_title())
@@ -332,9 +390,7 @@ class UITest(unittest.TestCase):
         self.assert_('risk' in s)
 
         # enable mint
-        for h in ui.handlers:
-            if h.module == 'mint':
-                h.enable()
+        ui.backend().set_enabled('kmod:mint', True)
         (h, s) = ui.main_window_text()
         self.assert_('Proprietary drivers are being used' in h)
         self.assert_('risk' in s)
@@ -342,12 +398,20 @@ class UITest(unittest.TestCase):
     def test_mode_free(self):
         '''calling with --mode free'''
 
-        sys.argv = ['ui-test', '--mode', 'free']
+        sys.argv = ['ui-test', '--list', '--mode', 'free']
         ui = sandbox.TestUI()
+        self.assertEqual(sandbox.TestUI().run(), 0)
+        displayed = set(ui.get_displayed_handlers())
         self.stop_capture()
 
-        self.assertEqual(set([h.module for h in ui.handlers]), 
-            set(['vanilla', 'firmwifi', 'foodmi']))
+        self.assertEqual(displayed, set(['kmod:vanilla',
+            'kmod:firmwifi', 'kmod:foodmi']))
+
+        self.assertEqual(len(self.stdout.strip().splitlines()), 3)
+        self.assert_('free module with available hardware, graphics card' in self.stdout)
+        self.assert_('kmod:vanilla ' in self.stdout)
+        self.assert_('kmod:firmwifi ' in self.stdout)
+        self.assert_('kmod:foodmi ' in self.stdout)
 
         self.failIf('Restricted' in ui.main_window_title())
 
@@ -358,12 +422,17 @@ class UITest(unittest.TestCase):
     def test_mode_nonfree(self):
         '''calling with --mode nonfree'''
 
-        sys.argv = ['ui-test', '--mode', 'nonfree']
+        sys.argv = ['ui-test', '--list', '--mode', 'nonfree']
         ui = sandbox.TestUI()
+        self.assertEqual(sandbox.TestUI().run(), 0)
+        displayed = set(ui.get_displayed_handlers())
         self.stop_capture()
 
-        self.assertEqual(set([h.module for h in ui.handlers]), 
-            set(['vanilla3d', 'mint']))
+        self.assertEqual(displayed, set(['kmod:mint', 'kmod:vanilla3d']))
+
+        self.assertEqual(len(self.stdout.strip().splitlines()), 2)
+        self.assert_('kmod:vanilla3d ' in self.stdout)
+        self.assert_('kmod:mint ' in self.stdout)
 
         self.assert_('Restricted' in ui.main_window_title())
 
@@ -390,8 +459,8 @@ class UITest(unittest.TestCase):
 
             # new free and nonfree drivers which are not enabled -> nonfree notification
             ui = sandbox.TestUI()
-            for h in ui.handlers:
-                h.disable()
+            for h in ui.backend().available():
+                ui.backend().set_enabled(h, False)
             self.assertEqual(ui.run(), 0)
 
             self.assertRaises(IndexError, ui.pop_error)
@@ -415,9 +484,9 @@ class UITest(unittest.TestCase):
 
             # enable the free drivers again and load them -> no notification
             ui.main_loop_active = False
-            for h in ui.handlers:
-                h.enable()
-                assert h.used()
+            for h in ui.backend().available('free'):
+                ui.backend().set_enabled(h, True)
+                assert jockey.ui.bool(ui.backend().handler_info(h)['used'])
             self.assertEqual(ui.run(), 1)
             self.assertRaises(IndexError, ui.pop_error)
             self.assertRaises(IndexError, ui.pop_notification)
@@ -427,9 +496,9 @@ class UITest(unittest.TestCase):
             # notification about usage
             sys.argv = ['ui-test', '--check', '-m', 'nonfree']
             ui = sandbox.TestUI()
-            for h in ui.handlers:
-                h.enable()
-                assert h.used()
+            for h in ui.backend().available('nonfree'):
+                ui.backend().set_enabled(h, True)
+                self.assert_(jockey.ui.bool(ui.backend().handler_info(h)['used']))
             self.assertEqual(ui.run(), 1)
             self.assertRaises(IndexError, ui.pop_error)
             self.assertEqual('New restricted drivers in use', ui.pop_notification()[0])
@@ -460,7 +529,7 @@ class UITest(unittest.TestCase):
         self.assertRaises(IndexError, ui.pop_error)
 
         # we got progress reports
-        self.assertEqual(ui.cur_download, ['file://' + self.big_file,
+        self.assertEqual(ui.cur_progress, ['file://' + self.big_file,
             len(self.big_file_contents), len(self.big_file_contents)])
 
         # specified file name
@@ -473,7 +542,7 @@ class UITest(unittest.TestCase):
         self.assertRaises(IndexError, ui.pop_error)
 
         # we got progress reports
-        self.assertEqual(ui.cur_download, ['file://' + self.big_file,
+        self.assertEqual(ui.cur_progress, ['file://' + self.big_file,
             len(self.big_file_contents), len(self.big_file_contents)])
 
         # nonexisting file
@@ -490,27 +559,27 @@ class UITest(unittest.TestCase):
         ui = sandbox.TestUI()
 
         # temporary file
-        ui.cancel_download = True
+        ui.cancel_progress = True
         (fname, h) = ui.download_url('file://' + self.big_file)
         self.assertEqual(fname, None)
 
         # we got progress reports
-        self.assertEqual(ui.cur_download[0], 'file://' + self.big_file)
-        self.assertEqual(ui.cur_download[2], len(self.big_file_contents))
-        ratio = float(ui.cur_download[1])/len(self.big_file_contents)
+        self.assertEqual(ui.cur_progress[0], 'file://' + self.big_file)
+        self.assertEqual(ui.cur_progress[2], len(self.big_file_contents))
+        ratio = float(ui.cur_progress[1])/len(self.big_file_contents)
         self.assert_(ratio >= 0.3, ratio)
         self.assert_(ratio < 0.5)
 
         # specified file name
         dest = os.path.join(OSLib.inst.workdir, 'destfile')
-        ui.cancel_download = True
+        ui.cancel_progress = True
         (fname, h) = ui.download_url('file://' + self.big_file, dest)
         self.assertEqual(fname, None)
 
         # we got progress reports
-        self.assertEqual(ui.cur_download[0], 'file://' + self.big_file)
-        self.assertEqual(ui.cur_download[2], len(self.big_file_contents))
-        ratio = float(ui.cur_download[1])/len(self.big_file_contents)
+        self.assertEqual(ui.cur_progress[0], 'file://' + self.big_file)
+        self.assertEqual(ui.cur_progress[2], len(self.big_file_contents))
+        ratio = float(ui.cur_progress[1])/len(self.big_file_contents)
         self.assert_(ratio >= 0.3, ratio)
         self.assert_(ratio < 0.5)
 
@@ -521,14 +590,30 @@ class UITest(unittest.TestCase):
         # fork test HTTP server
         pid = os.fork()
         if pid == 0:
-            os.chdir(OSLib.inst.workdir)
-            httpd = BaseHTTPServer.HTTPServer(('', 8427),
-                SimpleHTTPServer.SimpleHTTPRequestHandler)
-            httpd.handle_request()
-            os._exit(0)
+            try:
+                os.chdir(OSLib.inst.workdir)
+                httpd = BaseHTTPServer.HTTPServer(('', 8427),
+                    SimpleHTTPServer.SimpleHTTPRequestHandler)
+                # first request is for startup check, second for the actual test
+                httpd.handle_request()
+                httpd.handle_request()
+                os._exit(0)
+            except:
+                print '********** UI HTTP server failed: ***********'
+                traceback.print_exc()
+                os._exit(1)
         else:
             self.http_pid = pid
-            time.sleep(0.5) # TODO: bad race condition fix
+            # wait until server is ready
+            while True:
+                time.sleep(0.1)
+                try:
+                    f = urllib.urlopen('http://localhost:8427')
+                except IOError:
+                    continue
+                f.read()
+                f.close()
+                break
 
     def _join_http_server(self, kill=False):
         '''Wait for the forked HTTP server to finish and assert that it exits
@@ -536,10 +621,9 @@ class UITest(unittest.TestCase):
 
         if kill:
             os.kill(self.http_pid, signal.SIGTERM)
-
-        # TODO: check exit code
-        #self.assertEqual(os.wait()[1], 0)
-        os.wait()
+            os.wait() # undefined exit code
+        else:
+            self.assertEqual(os.wait(), (self.http_pid, 0))
 
     def test_download_http_nocancel(self):
         '''download_url(), HTTP, no cancelling'''
@@ -556,7 +640,7 @@ class UITest(unittest.TestCase):
         self.assertRaises(IndexError, ui.pop_error)
 
         # we got progress reports
-        self.assertEqual(ui.cur_download, ['http://localhost:8427/stuff',
+        self.assertEqual(ui.cur_progress, ['http://localhost:8427/stuff',
             len(self.big_file_contents), len(self.big_file_contents)])
 
         # specified file name
@@ -571,7 +655,7 @@ class UITest(unittest.TestCase):
         self.assertRaises(IndexError, ui.pop_error)
 
         # we got progress reports
-        self.assertEqual(ui.cur_download, ['http://localhost:8427/stuff',
+        self.assertEqual(ui.cur_progress, ['http://localhost:8427/stuff',
             len(self.big_file_contents), len(self.big_file_contents)])
 
         # nonexisting file
@@ -601,7 +685,7 @@ class UITest(unittest.TestCase):
 
         # temporary file
         self._fork_http_server()
-        ui.cancel_download = True
+        ui.cancel_progress = True
         (fname, h) = ui.download_url('http://localhost:8427/stuff')
         self._join_http_server()
         self.assertEqual(fname, None)
@@ -609,14 +693,14 @@ class UITest(unittest.TestCase):
         self.assertRaises(IndexError, ui.pop_error)
 
         # we got progress reports
-        self.assertEqual(ui.cur_download[2], len(self.big_file_contents))
-        ratio = float(ui.cur_download[1])/len(self.big_file_contents)
+        self.assertEqual(ui.cur_progress[2], len(self.big_file_contents))
+        ratio = float(ui.cur_progress[1])/len(self.big_file_contents)
         self.assert_(ratio >= 0.3, ratio)
         self.assert_(ratio < 0.5)
 
         # specified file name
         dest = os.path.join(OSLib.inst.workdir, 'destfile')
-        ui.cancel_download = True
+        ui.cancel_progress = True
         self._fork_http_server()
         (fname, h) = ui.download_url('http://localhost:8427/stuff', dest)
         self._join_http_server()
@@ -625,46 +709,11 @@ class UITest(unittest.TestCase):
         self.assertRaises(IndexError, ui.pop_error)
 
         # we got progress reports
-        self.assertEqual(ui.cur_download[2], len(self.big_file_contents))
-        ratio = float(ui.cur_download[1])/len(self.big_file_contents)
+        self.assertEqual(ui.cur_progress[2], len(self.big_file_contents))
+        ratio = float(ui.cur_progress[1])/len(self.big_file_contents)
         self.assert_(ratio >= 0.3, ratio)
         self.assert_(ratio < 0.5)
 
-    def test_install_packages(self):
-        '''{install,remove}_package() and installed_packages logging'''
-
-        pkglog = os.path.join(OSLib.inst.backup_dir, 'installed_packages')
-        ui = sandbox.TestUI()
-        self.failIf(os.path.exists(pkglog))
-
-        # coreutils package is already installed
-        self.assert_(OSLib.inst.package_installed('coreutils'))
-        ui.install_package('coreutils')
-        self.failIf(os.path.exists(pkglog))
-
-        # install mesa-vanilla
-        self.failIf(OSLib.inst.package_installed('mesa-vanilla'))
-        ui.install_package('mesa-vanilla')
-        self.assert_(OSLib.inst.package_installed('mesa-vanilla'))
-        self.assertEqual(open(pkglog).read(), 'mesa-vanilla\n')
-
-        # install pretzel
-        self.failIf(OSLib.inst.package_installed('pretzel'))
-        ui.install_package('pretzel')
-        self.assert_(OSLib.inst.package_installed('pretzel'))
-        inst = sorted(open(pkglog).readlines())
-        self.assertEqual(inst, ['mesa-vanilla\n', 'pretzel\n'])
-
-        # remove mesa-vanilla
-        ui.remove_package('mesa-vanilla')
-        self.failIf(OSLib.inst.package_installed('mesa-vanilla'))
-        self.assertEqual(open(pkglog).read(), 'pretzel\n')
-
-        # remove pretzel
-        ui.remove_package('pretzel')
-        self.failIf(OSLib.inst.package_installed('pretzel'))
-        self.failIf(os.path.exists(pkglog))
-
     def test_check_composite_noavail(self):
         '''calling with --check-composite and no available driver'''
 
@@ -674,13 +723,12 @@ class UITest(unittest.TestCase):
         h = XorgDriverHandler(ui, 'vanilla3d', 'mesa-vanilla', 'v3d',
             'vanilla')
         self.failIf(h.enabled())
-        ui.handlers.add(h)
+        ui.backend().handlers[h.id()] = h
 
         self.assertEqual(ui.run(), 1)
         self.stop_capture()
 
         self.assert_('no available' in self.stderr)
-        self.failIf(os.path.exists(self.open_app_log))
 
     def test_check_composite_avail(self):
         '''calling with --check-composite and available driver'''
@@ -692,18 +740,19 @@ class UITest(unittest.TestCase):
             def __init__(self, ui):
                 XorgDriverHandler.__init__(self, ui, 'vanilla3d',
                     'mesa-vanilla', 'v3d', 'vanilla')
-            def supports_composite(self):
+            def enables_composite(self):
                 return True
 
-        h = XorgComp(ui)
+        h = XorgComp(ui.backend())
+        h_id = h.id()
         self.failIf(h.enabled())
-        ui.handlers.add(h)
+        ui.backend().handlers[h_id] = h
 
+        ui.confirm_response = True
         self.assertEqual(ui.run(), 0)
 
         self.stop_capture()
-        self.assertEqual(open(self.open_app_log).read(), 
-            '--confirm --enable kmod:vanilla3d\n')
+        self.assert_(jockey.ui.bool(ui.backend().handler_info(h_id)['enabled']))
 
     def test_check_composite_enabled(self):
         '''calling with --check-composite and already enabled driver'''
@@ -715,16 +764,138 @@ class UITest(unittest.TestCase):
             def __init__(self, ui):
                 XorgDriverHandler.__init__(self, ui, 'vanilla3d',
                     'mesa-vanilla', 'v3d', 'vanilla')
-            def supports_composite(self):
+            def enables_composite(self):
                 return True
             def enabled(self):
                 return True
 
         h = XorgComp(ui)
-        ui.handlers.add(h)
+        ui.backend().handlers[h.id()] = h
 
+        # note: do not set confirm_response here, shouldn't ask
         self.assertEqual(ui.run(), 1)
 
         self.stop_capture()
-        self.failIf(os.path.exists(self.open_app_log))
-        self.assert_('already enabled' in self.stderr)
+        self.assert_('already supports' in self.stderr, self.stderr)
+
+    # TODO: disabled, this currently breaks
+    def teest_ui_dbus(self):
+        '''UI --dbus-server'''
+
+        svr_pid = os.fork()
+        if svr_pid == 0:
+            sys.argv = ['ui-test', '--dbus-server']
+            try:
+                sandbox.TestUI().run()
+            except:
+                print '********** UI D-BUS server failed: ***********'
+                traceback.print_exc()
+                os._exit(1)
+            os._exit(0)
+
+        time.sleep(2)
+        iface = sandbox.TestUI.get_dbus_client()
+
+        self.assertEqual(iface.search_driver('unknown:foo'), False)
+
+        # shuts down after one request
+        try:
+            iface.search_driver('unknown:bar')
+            fail('should only accept one method call and then terminate')
+        except Exception, e:
+            self.assertEqual(e._dbus_error_name, 'org.freedesktop.DBus.Error.NoReply')
+
+        # give the daemon a second to terminate
+        timeout = 10
+        while timeout >= 0:
+            (pid, exitcode) = os.waitpid(svr_pid, os.WNOHANG)
+            if pid > 0:
+                self.assertEqual(svr_pid, pid)
+                self.assertEqual(exitcode, 0)
+                break
+            time.sleep(0.1)
+            timeout -= 1
+        if timeout <= 0:
+            os.kill(svr_pid, signal.SIGKILL)
+            os.waitpid(svr_pid, 0)
+            self.fail('UI D-BUS server does not auto-terminate, killing')
+
+    def test_search_driver_errors(self):
+        '''D-BUS API: Search for drivers which cannot be enabled'''
+
+        ui = sandbox.TestUI()
+        self.stop_capture()
+
+        # does not exist
+        self.assertEqual(ui.search_driver('unknown:foo'), (False, []))
+        self.assertEqual(ui.get_displayed_handlers(), [])
+
+        # kmod:vanilla, already enabled
+        self.assertEqual(ui.search_driver(
+            'modalias:pci:v00001001d0sv0sd0bc03sc00i00'), (False, []))
+        self.assertEqual(ui.get_displayed_handlers(), ['kmod:vanilla'])
+
+    def test_search_driver_disabled(self):
+        '''D-BUS API: Search for a noninstalled driver'''
+
+        # search_driver() should not display available==True handlers
+        open (os.path.join(OSLib.inst.handler_dir, 'kmod_avail.py'), 'w').write(
+            sandbox.h_availability_py)
+        ui = sandbox.TestUI()
+
+        self.assert_('kmod:spam' not in ui.backend().available())
+
+        sandbox.start_driverdb_server()
+        try:
+            ui.backend().add_driverdb('XMLRPCDriverDB', ['http://localhost:80080'])
+            # first match is kmod:mint:UbuOne in our fake db
+            result = ui.search_driver('pci:9876/FEDC')
+        finally:
+            sandbox.stop_driverdb_server()
+
+        self.stop_capture()
+        self.assertEqual(result, (False, []))
+        self.assertEqual(ui.get_displayed_handlers(), ['kmod:spam'])
+
+        info = ui.backend().handler_info('kmod:spam')
+        self.assertEqual(info['enabled'], 'True')
+
+    def test_search_driver_enable(self):
+        '''D-BUS API: Search for a driver which gets enabled'''
+
+        ui = sandbox.TestUI()
+        self.stop_capture()
+
+        ui.backend().handlers['kmod:vanilla3d'].package = 'mesa-vanilla'
+
+        # don't enable the driver
+        result = ui.search_driver('modalias:pci:v00001001d00003D0sv0sd0bc03sc0i0')
+        self.assertEqual(set(ui.get_displayed_handlers()),
+            set(['kmod:vanilla3d', 'kmod:vanilla']))
+        self.assertEqual(result, (False, []))
+        self.assertEqual(ui.backend().handler_info('kmod:vanilla3d')['enabled'],
+            'False')
+
+        # enable it now
+        ui.ui_main_loop = lambda: ui.set_handler_enable('kmod:vanilla3d',
+            'enable', False, False) # fake main loop to enable kmod:vanilla3d
+        result = ui.search_driver('modalias:pci:v00001001d00003D0sv0sd0bc03sc0i0')
+        self.assertEqual(set(ui.get_displayed_handlers()),
+            set(['kmod:vanilla3d', 'kmod:vanilla']))
+        self.assertEqual(result, (True, sandbox.fake_pkginfo['default']['mesa-vanilla']['files']))
+        self.assertEqual(ui.backend().handler_info('kmod:vanilla3d')['enabled'],
+            'True')
+
+    def test_hwid_to_display_string(self):
+        '''hwid_to_display_string()'''
+
+        ui = sandbox.TestUI()
+        self.stop_capture()
+
+        self.assertEqual(ui.hwid_to_display_string('foo'), 'foo')
+        self.assertEqual(ui.hwid_to_display_string('foo:bar'), 'foo:bar')
+        self.assertEqual(ui.hwid_to_display_string('printer_deviceid:MFG:Samsung;MDL:ML-1610;CMD:GDI'), 
+            'Samsung ML-1610')
+        self.assertEqual(ui.hwid_to_display_string('printer_deviceid:MDL:InkShot 12;CMD:GDI;MFG:PH'), 
+            'PH InkShot 12')
+
